#include "string.h"

#include "nrf.h"
#include "nrf_uarte.h"
#include "nrf_gpio.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#define DMA_MAX_LEN 256

// Global receive buffer and count of bytes in it
uint8_t gpBuffer[128];
size_t gBytesUsed = 0;

// The interrupt handler: only handles Rx data as Tx is blocking.
static void irqHandler(NRF_UARTE_Type *pReg)
{
    size_t length;
    bool wasRxEvent = false;

    if (nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_ERROR)) {
        // Clear any errors
        nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_ERROR);
        nrf_uarte_errorsrc_get_and_clear(pReg);
        wasRxEvent = true;
    } else if (nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_ENDRX)) {
        // Clear Rx event and grab bytes used
        nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_ENDRX);
        gBytesUsed += nrf_uarte_rx_amount_get(pReg);
        wasRxEvent = true;
    }

    if (nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_RXTO)) {
        // Clear time-out event
        nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_RXTO);
    }

    if (nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_ENDTX)) {
        nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_ENDTX);
        // Transmitter has to be stopped by triggering STOPTX task to achieve
        // the lowest possible level of the UARTE power consumption.
        nrf_uarte_task_trigger(pReg, NRF_UARTE_TASK_STOPTX);
    }

    if (nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_TXSTOPPED)) {
        nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_TXSTOPPED);
    }

    if (wasRxEvent) {
        // Start the next receive
        length = sizeof(gpBuffer) - gBytesUsed;
        if (length > DMA_MAX_LEN) {
            length = DMA_MAX_LEN;
        }
        nrf_uarte_rx_buffer_set(pReg,
                                gpBuffer + gBytesUsed,
                                length);
        nrf_uarte_task_trigger(pReg, NRF_UARTE_TASK_STARTRX);
    }
}

// Derived from the NRFX functions nrfx_is_in_ram() and
// nrfx_is_word_aligned(), check if a buffer pointer is good
// for DMA
__STATIC_INLINE bool isGoodForDma(void const *pBuffer)
{
    return (((((uint32_t) pBuffer) & 0x3u) == 0u) && 
            ((((uint32_t) pBuffer) & 0xE0000000u) == 0x20000000u));
}

// Derived from the NRFX function nrfx_get_irq_number()
__STATIC_INLINE IRQn_Type getIrqNumber(void const *pReg)
{
    return (IRQn_Type) (uint8_t)((uint32_t)(pReg) >> 12);
}

/* ----------------------------------------------------------------
 * INTERRUPT HANDLERS
 * -------------------------------------------------------------- */

static void UARTE0_UART0_IRQHandler(void)
{
    irqHandler(NRF_UARTE0);
}

static void UARTE1_IRQHandler(void)
{
    irqHandler(NRF_UARTE1);
}

/* ----------------------------------------------------------------
 * UARTE FUNCTIONS
 * -------------------------------------------------------------- */

// Initialise a UARTE.
void uarteInit()
{
    int32_t pinTx = 22;
    int32_t pinRx = 23;
    NRF_UARTE_Type *pReg = NRF_UARTE0;

    // Set baud rate
    nrf_uarte_baudrate_set(pReg, UARTE_BAUDRATE_BAUDRATE_Baud115200);

    // Set Tx/Rx pins
    nrf_gpio_pin_set(pinTx);
    nrf_gpio_cfg_output(pinTx);
    nrf_gpio_cfg_input(pinRx, NRF_GPIO_PIN_NOPULL);
    nrf_uarte_txrx_pins_set(pReg, pinTx, pinRx);

    // Set flow control off
    nrf_uarte_configure(pReg, NRF_UARTE_PARITY_EXCLUDED, NRF_UARTE_HWFC_DISABLED);

    // Set interrupts and buffer and let it go
    nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_ENDRX);
    nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_ENDTX);
    nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_ERROR);
    nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_RXTO);
    nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_TXSTOPPED);
    nrf_uarte_rx_buffer_set(pReg, gpBuffer, DMA_MAX_LEN);
    nrf_uarte_task_trigger(pReg, NRF_UARTE_TASK_STARTRX);
    nrf_uarte_int_enable(pReg, NRF_UARTE_INT_ENDRX_MASK |
                               NRF_UARTE_INT_ENDTX_MASK |
                               NRF_UARTE_INT_ERROR_MASK |
                               NRF_UARTE_INT_RXTO_MASK  |
                               NRF_UARTE_INT_TXSTOPPED_MASK);
    NRFX_IRQ_PRIORITY_SET(getIrqNumber((void *) pReg),
                          NRFX_UARTE_DEFAULT_CONFIG_IRQ_PRIORITY);
    NRFX_IRQ_ENABLE(getIrqNumber((void *) (pReg)));
}

// Shutdown a UART.
void uarteDeinit()
{
    NRF_UARTE_Type *pReg = NRF_UARTE0;

    // Disable interrupts
    nrf_uarte_int_disable(pReg, NRF_UARTE_INT_ENDRX_MASK |
                                NRF_UARTE_INT_ENDTX_MASK |
                                NRF_UARTE_INT_ERROR_MASK |
                                NRF_UARTE_INT_RXTO_MASK  |
                                NRF_UARTE_INT_TXSTOPPED_MASK);
    NRFX_IRQ_DISABLE(nrfx_get_irq_number((void *) (pReg)));

    // Make sure all transfers are finished before UARTE is
    // disabled to achieve the lowest power consumption
    nrf_uarte_shorts_disable(pReg, NRF_UARTE_SHORT_ENDRX_STARTRX);
    nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_RXTO);
    nrf_uarte_task_trigger(pReg, NRF_UARTE_TASK_STOPRX);
    nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_TXSTOPPED);
    nrf_uarte_task_trigger(pReg, NRF_UARTE_TASK_STOPTX);
    while (!nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_TXSTOPPED) ||
           !nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_RXTO)) {}

    // Disable the UARTE
    nrf_uarte_disable(pReg);

    // Put the pins back
    nrf_gpio_cfg_default(nrf_uarte_tx_pin_get(pReg));
    nrf_gpio_cfg_default(nrf_uarte_rx_pin_get(pReg));
    nrf_uarte_txrx_pins_disconnect(pReg);
}

// Read from the UARTE interface.
int32_t uarteRead(uint8_t *pBuffer, size_t sizeBytes)
{
    size_t size = 0;

    if (pBuffer != NULL) {
        // Copy out the data
        size = gBytesUsed;
        if (size > sizeBytes) {
            size = sizeBytes;
        }
        memcpy(pBuffer, gpBuffer, size);
        gBytesUsed -= size;
    }

    return (int32_t) size;
}

// Blocking write to the UARTE interface.
int32_t uarteWrite(const uint8_t *pBuffer,
                   size_t sizeBytes)
{
    int32_t size = 0;
    const uint8_t *pTxBuffer = NULL;
    uint8_t *pTmpBuffer = NULL;
    NRF_UARTE_Type *pReg = NRF_UARTE0;

    if (pBuffer != NULL) {
        // If buffer is not good for DMA (e.g. if it
        // is in flash) then copy to RAM.
        if (!isGoodForDma(pBuffer)) {
            pTmpBuffer = (uint8_t *) malloc(sizeBytes);
            if (pTmpBuffer != NULL) {
                memcpy(pTmpBuffer, pBuffer, sizeBytes);
                pTxBuffer = pTmpBuffer;
            }
        } else {
            pTxBuffer = pBuffer;
        }

        if (pTxBuffer != NULL) {
            // Set up the flags
            nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_ENDTX);
            nrf_uarte_event_clear(pReg, NRF_UARTE_EVENT_TXSTOPPED);
            nrf_uarte_tx_buffer_set(pReg, pTxBuffer, sizeBytes);
            nrf_uarte_task_trigger(pReg, NRF_UARTE_TASK_STARTTX);

            // Wait for the transmission to complete
            // GETS STUCK HERE!!!
            while (!nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_ENDTX)) {}

            // Put UARTE into lowest power state.
            nrf_uarte_task_trigger(pReg, NRF_UARTE_TASK_STOPTX);
            while (!nrf_uarte_event_check(pReg, NRF_UARTE_EVENT_TXSTOPPED)) {}

            size = sizeBytes;
        }

        free(pTmpBuffer);
    }

    return (int32_t) size;
}

/* ----------------------------------------------------------------
 * MAIN
 * -------------------------------------------------------------- */

// Entry point
int main(void)
{
    uint8_t txBuffer[] = "Hello\n";
    size_t txBufferLength = strlen(txBuffer);

    // Set up logging
    NRF_LOG_INIT(NULL);
    NRF_LOG_DEFAULT_BACKENDS_INIT();

    // Init UART 0
    uarteInit();

    // Write the tx Buffer to UART 0
    uarteWrite(txBuffer, txBufferLength);

    // Deinit UART 0
    uarteDeinit();

    while(1) {};

    return 0;
}